@@ -20,13 +20,13 @@
  * SECTION:kgx-application
  * @title: KgxApplication
  * @short_description: Application
- * 
+ *
  * The application, on the face of it nothing particularly interesting but
  * under the hood it contains a #GSource used to monitor the shells (and
  * there children) running in the open #KgxWindow s
  */
 
-#define G_LOG_DOMAIN "Kgx"
+#include "kgx-config.h"
 
 #include <glib/gi18n.h>
 #include <vte/vte.h>
@@ -35,9 +35,10 @@
 
 #include "rgba.h"
 
-#include "kgx-config.h"
 #include "kgx-application.h"
 #include "kgx-window.h"
+#include "kgx-pages.h"
+#include "kgx-simple-tab.h"
 
 #define LOGO_COL_SIZE 28
 #define LOGO_ROW_SIZE 14
@@ -49,11 +50,13 @@ enum {
   PROP_THEME,
   PROP_FONT,
   PROP_FONT_SCALE,
+  PROP_SCROLLBACK_LINES,
   LAST_PROP
 };
 
 static GParamSpec *pspecs[LAST_PROP] = { NULL, };
 
+
 static void
 kgx_application_set_theme (KgxApplication *self,
                            KgxTheme        theme)
@@ -65,17 +68,28 @@ kgx_application_set_theme (KgxApplication *self,
   g_object_notify_by_pspec (G_OBJECT (self), pspecs[PROP_THEME]);
 }
 
+
 static void
 kgx_application_set_scale (KgxApplication *self,
                            gdouble         scale)
 {
+  GAction *action;
+
   g_return_if_fail (KGX_IS_APPLICATION (self));
 
-  self->scale = scale;
+  self->scale = CLAMP (scale, 0.5, 2.0);
+
+  action = g_action_map_lookup_action (G_ACTION_MAP (self), "zoom-out");
+  g_simple_action_set_enabled (G_SIMPLE_ACTION (action), self->scale > 0.5);
+  action = g_action_map_lookup_action (G_ACTION_MAP (self), "zoom-normal");
+  g_simple_action_set_enabled (G_SIMPLE_ACTION (action), self->scale != 1.0);
+  action = g_action_map_lookup_action (G_ACTION_MAP (self), "zoom-in");
+  g_simple_action_set_enabled (G_SIMPLE_ACTION (action), self->scale < 2.0);
 
   g_object_notify_by_pspec (G_OBJECT (self), pspecs[PROP_FONT_SCALE]);
 }
 
+
 static void
 kgx_application_set_property (GObject      *object,
                               guint         property_id,
@@ -91,6 +105,9 @@ kgx_application_set_property (GObject      *object,
     case PROP_FONT_SCALE:
       kgx_application_set_scale (self, g_value_get_double (value));
       break;
+    case PROP_SCROLLBACK_LINES:
+      self->scrollback_lines = g_value_get_int64 (value);
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
       break;
@@ -115,25 +132,32 @@ kgx_application_get_property (GObject    *object,
     case PROP_FONT_SCALE:
       g_value_set_double (value, self->scale);
       break;
+    case PROP_SCROLLBACK_LINES:
+      g_value_set_int64 (value, self->scrollback_lines);
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
       break;
   }
 }
 
+
 static void
 kgx_application_finalize (GObject *object)
 {
   KgxApplication *self = KGX_APPLICATION (object);
 
+  g_clear_object (&self->settings);
   g_clear_object (&self->desktop_interface);
 
   g_clear_pointer (&self->watching, g_tree_unref);
   g_clear_pointer (&self->children, g_tree_unref);
+  g_clear_pointer (&self->pages, g_tree_unref);
 
   G_OBJECT_CLASS (kgx_application_parent_class)->finalize (object);
 }
 
+
 static void
 kgx_application_activate (GApplication *app)
 {
@@ -145,16 +169,20 @@ kgx_application_activate (GApplication *app)
   /* Get the current window or create one if necessary. */
   window = gtk_application_get_active_window (GTK_APPLICATION (app));
   if (window == NULL) {
-    window = g_object_new (KGX_TYPE_WINDOW,
-                           "application", app,
-                           "close-on-zero", TRUE,
-                           NULL);
+    kgx_application_add_terminal (KGX_APPLICATION (app),
+                                  NULL,
+                                  timestamp,
+                                  NULL,
+                                  NULL,
+                                  NULL);
+
+    return;
   }
 
   gtk_window_present_with_time (window, timestamp);
 }
 
-#if HAS_GTOP
+
 static gboolean
 handle_watch_iter (gpointer pid,
                    gpointer val,
@@ -174,24 +202,26 @@ handle_watch_iter (gpointer pid,
       struct ProcessWatch *child_watch = g_new (struct ProcessWatch, 1);
 
       child_watch->process = g_rc_box_acquire (process);
-      child_watch->window = g_object_ref (watch->window);
+      child_watch->page = g_object_ref (watch->page);
 
       g_debug ("Hello %i!", GPOINTER_TO_INT (pid));
 
       g_tree_insert (self->children, pid, child_watch);
     }
 
-    kgx_window_push_child (watch->window, process);
+    kgx_tab_push_child (watch->page, process);
   }
 
   return FALSE;
 }
 
+
 struct RemoveDead {
   GTree     *plist;
   GPtrArray *dead;
 };
 
+
 static gboolean
 remove_dead (gpointer pid,
              gpointer val,
@@ -203,7 +233,7 @@ remove_dead (gpointer pid,
   if (!g_tree_lookup (data->plist, pid)) {
     g_debug ("%i marked as dead", GPOINTER_TO_INT (pid));
 
-    kgx_window_pop_child (watch->window, watch->process);
+    kgx_tab_pop_child (watch->page, watch->process);
 
     g_ptr_array_add (data->dead, pid);
   }
@@ -211,6 +241,7 @@ remove_dead (gpointer pid,
   return FALSE;
 }
 
+
 static gboolean
 watch (gpointer data)
 {
@@ -236,43 +267,47 @@ watch (gpointer data)
 
   return G_SOURCE_CONTINUE;
 }
-#endif
 
 static inline void
 set_watcher (KgxApplication *self, gboolean focused)
 {
   g_debug ("updated watcher focused? %s", focused ? "yes" : "no");
 
-  #if HAS_GTOP
   if (self->timeout != 0) {
     g_source_remove (self->timeout);
   }
 
   // Slow down polling when nothing is focused
   self->timeout = g_timeout_add (focused ? 500 : 2000, watch, self);
-  // Translators: This is the name of the timeout that looks for programs
-  // running in the terminal
-  g_source_set_name_by_id (self->timeout, _("child watcher"));
-  #endif
+  g_source_set_name_by_id (self->timeout, "[kgx] child watcher");
 }
 
+
 static void
 kgx_application_startup (GApplication *app)
 {
+  KgxApplication *self = KGX_APPLICATION (app);
   GtkSettings    *gtk_settings;
-  GSettings      *settings;
   GtkCssProvider *provider;
   const char *const new_window_accels[] = { "<shift><primary>n", NULL };
+  const char *const new_tab_accels[] = { "<shift><primary>t", NULL };
+  const char *const close_tab_accels[] = { "<shift><primary>w", NULL };
   const char *const copy_accels[] = { "<shift><primary>c", NULL };
   const char *const paste_accels[] = { "<shift><primary>v", NULL };
   const char *const find_accels[] = { "<shift><primary>f", NULL };
   const char *const zoom_in_accels[] = { "<primary>plus", NULL };
   const char *const zoom_out_accels[] = { "<primary>minus", NULL };
+  const char *const zoom_normal_accels[] = { "<primary>0", NULL };
+
+  g_set_prgname (g_application_get_application_id (app));
 
   g_type_ensure (KGX_TYPE_TERMINAL);
+  g_type_ensure (KGX_TYPE_PAGES);
 
   G_APPLICATION_CLASS (kgx_application_parent_class)->startup (app);
 
+  hdy_init ();
+
   gtk_settings = gtk_settings_get_default ();
 
   g_object_set (G_OBJECT (gtk_settings),
@@ -281,6 +316,10 @@ kgx_application_startup (GApplication *app)
 
   gtk_application_set_accels_for_action (GTK_APPLICATION (app),
                                          "win.new-window", new_window_accels);
+  gtk_application_set_accels_for_action (GTK_APPLICATION (app),
+                                         "win.new-tab", new_tab_accels);
+  gtk_application_set_accels_for_action (GTK_APPLICATION (app),
+                                         "win.close-tab", close_tab_accels);
   gtk_application_set_accels_for_action (GTK_APPLICATION (app),
                                          "term.copy", copy_accels);
   gtk_application_set_accels_for_action (GTK_APPLICATION (app),
@@ -288,16 +327,19 @@ kgx_application_startup (GApplication *app)
   gtk_application_set_accels_for_action (GTK_APPLICATION (app),
                                          "win.find", find_accels);
   gtk_application_set_accels_for_action (GTK_APPLICATION (app),
-                                         "win.zoom-in", zoom_in_accels);
+                                         "app.zoom-in", zoom_in_accels);
   gtk_application_set_accels_for_action (GTK_APPLICATION (app),
-                                         "win.zoom-out", zoom_out_accels);
+                                         "app.zoom-out", zoom_out_accels);
+  gtk_application_set_accels_for_action (GTK_APPLICATION (app),
+                                         "app.zoom-normal", zoom_normal_accels);
 
-  settings = g_settings_new ("org.gnome.zbrown.KingsCross");
-  g_settings_bind (settings, "theme", app, "theme", G_SETTINGS_BIND_DEFAULT);
-  g_settings_bind (settings, "font-scale", app, "font-scale", G_SETTINGS_BIND_DEFAULT);
+  self->settings = g_settings_new (KGX_APPLICATION_ID);
+  g_settings_bind (self->settings, "theme", app, "theme", G_SETTINGS_BIND_DEFAULT);
+  g_settings_bind (self->settings, "font-scale", app, "font-scale", G_SETTINGS_BIND_DEFAULT);
+  g_settings_bind (self->settings, "scrollback-lines", app, "scrollback-lines", G_SETTINGS_BIND_DEFAULT);
 
   provider = gtk_css_provider_new ();
-  gtk_css_provider_load_from_resource (provider, RES_PATH "styles.css");
+  gtk_css_provider_load_from_resource (provider, KGX_APPLICATION_PATH "styles.css");
   gtk_style_context_add_provider_for_screen (gdk_screen_get_default (),
                                              GTK_STYLE_PROVIDER (provider),
                                              /* Is this stupid? Yes
@@ -308,36 +350,84 @@ kgx_application_startup (GApplication *app)
   set_watcher (KGX_APPLICATION (app), TRUE);
 }
 
+
+static void
+kgx_application_open (GApplication  *app,
+                      GFile        **files,
+                      int            n_files,
+                      const char    *hint)
+{
+  guint32 timestamp = GDK_CURRENT_TIME;
+
+  for (int i = 0; i < n_files; i++) {
+    kgx_application_add_terminal (KGX_APPLICATION (app),
+                                  NULL,
+                                  timestamp,
+                                  files[i],
+                                  NULL,
+                                  NULL);
+  }
+}
+
+
 static int
 kgx_application_command_line (GApplication            *app,
                               GApplicationCommandLine *cli)
 {
+  KgxApplication *self = KGX_APPLICATION (app);
+  guint32 timestamp = GDK_CURRENT_TIME;
   GVariantDict *options = NULL;
   const char *working_dir = NULL;
+  const char *title = NULL;
   const char *command = NULL;
-  GtkWidget *window;
-  g_autofree char *abs_path = NULL;
+  const char *const *shell = NULL;
+  const char *cwd = NULL;
+  gint64 scrollback;
+  gboolean tab;
+  g_autoptr (GFile) path = NULL;
 
   options = g_application_command_line_get_options_dict (cli);
+  cwd = g_application_command_line_get_cwd (cli);
 
   g_variant_dict_lookup (options, "working-directory", "^&ay", &working_dir);
+  g_variant_dict_lookup (options, "title", "&s", &title);
   g_variant_dict_lookup (options, "command", "^&ay", &command);
 
-  if (working_dir != NULL) {
-    abs_path = g_canonicalize_filename (working_dir, NULL);
+  if (g_variant_dict_lookup (options, "set-shell", "^as", &shell) && shell) {
+    g_settings_set_strv (self->settings, "shell", shell);
+
+    return EXIT_SUCCESS;
   }
 
-  window = g_object_new (KGX_TYPE_WINDOW,
-                         "application", app,
-                         "close-on-zero", command == NULL,
-                         "initial-work-dir", abs_path,
-                         "command", command,
-                         NULL);
-  gtk_widget_show (window);
+  if (g_variant_dict_lookup (options, "set-scrollback", "x", &scrollback)) {
+    g_settings_set_int64 (self->settings, "scrollback-lines", scrollback);
 
-  return 0;
+    return EXIT_SUCCESS;
+  }
+
+  if (working_dir != NULL) {
+    path = g_file_new_for_commandline_arg_and_cwd (working_dir, cwd);
+  }
+
+  if (path == NULL) {
+    path = g_file_new_for_path (cwd);
+  }
+
+  if (g_variant_dict_lookup (options, "tab", "b", &tab) && tab) {
+    kgx_application_add_terminal (self,
+                                  KGX_WINDOW (gtk_application_get_active_window (GTK_APPLICATION (self))),
+                                  timestamp,
+                                  path,
+                                  command,
+                                  title);
+  } else {
+    kgx_application_add_terminal (self, NULL, timestamp, path, command, title);
+  }
+
+  return EXIT_SUCCESS;
 }
 
+
 static void
 print_center (char *msg, int ign, short width)
 {
@@ -362,7 +452,7 @@ print_logo (short width)
   int i = 0;
   int half_screen = width / 2;
 
-  logo = g_file_new_for_uri ("resource:/" RES_PATH "logo.txt");
+  logo = g_file_new_for_uri ("resource:/" KGX_APPLICATION_PATH "logo.txt");
 
   g_file_load_contents (logo, NULL, &logo_text, NULL, NULL, &error);
 
@@ -392,21 +482,21 @@ kgx_application_handle_local_options (GApplication *app,
   if (g_variant_dict_lookup (options, "version", "b", &version)) {
     if (version) {
       // Translators: The leading # is intentional, the initial %s is the
-      // version of King's Cross itself, the latter format is the VTE version
-      g_print (_("# King’s Cross %s using VTE %u.%u.%u %s\n"),
+      // version of KGX itself, the latter format is the VTE version
+      g_print (_("# KGX %s using VTE %u.%u.%u %s\n"),
                PACKAGE_VERSION,
                vte_get_major_version (),
                vte_get_minor_version (),
                vte_get_micro_version (),
                vte_get_features ());
-      return 0;
+      return EXIT_SUCCESS;
     }
   }
 
   if (g_variant_dict_lookup (options, "about", "b", &about)) {
     if (about) {
-      g_autofree char *copyright = g_strdup_printf (_("Copyright © %s Zander Brown"),
-                                                    "2019");
+      g_autofree char *copyright = g_strdup_printf (_("© %s Zander Brown"),
+                                                    "2019-2021");
       struct winsize w;
       int padding = 0;
 
@@ -414,7 +504,7 @@ kgx_application_handle_local_options (GApplication *app,
 
       padding = ((w.ws_row -1) - (LOGO_ROW_SIZE + 5)) / 2;
 
-      for (int i = 0; i < padding - 1; i++) {
+      for (int i = 0; i < padding; i++) {
         g_print ("\n");
       }
 
@@ -423,19 +513,20 @@ kgx_application_handle_local_options (GApplication *app,
       print_center (PACKAGE_VERSION, -1, w.ws_col);
       print_center (_("Terminal Emulator"), -1, w.ws_col);
       print_center (copyright, -1, w.ws_col);
-      print_center ("GPL 3+\n", -1, w.ws_col);
+      print_center (_("GPL 3.0 or later"), -1, w.ws_col);
 
       for (int i = 0; i < padding; i++) {
         g_print ("\n");
       }
 
-      return 0;
+      return EXIT_SUCCESS;
     }
   }
 
   return G_APPLICATION_CLASS (kgx_application_parent_class)->handle_local_options (app, options);
 }
 
+
 static void
 kgx_application_class_init (KgxApplicationClass *klass)
 {
@@ -448,18 +539,19 @@ kgx_application_class_init (KgxApplicationClass *klass)
 
   app_class->activate = kgx_application_activate;
   app_class->startup = kgx_application_startup;
+  app_class->open = kgx_application_open;
   app_class->command_line = kgx_application_command_line;
   app_class->handle_local_options = kgx_application_handle_local_options;
 
   /**
    * KgxApplication:theme:
-   * 
+   *
    * The palette to use, one of the values of #KgxTheme
-   * 
+   *
    * Officially only "night" exists, "hacker" is just a little fun
-   * 
-   * Bound to /org/gnome/zbrown/KingsCross/theme so changes persist
-   * 
+   *
+   * Bound to ‘theme’ GSetting so changes persist
+   *
    * Stability: Private
    */
   pspecs[PROP_THEME] =
@@ -477,20 +569,38 @@ kgx_application_class_init (KgxApplicationClass *klass)
                          0.5, 2.0, 1.0,
                          G_PARAM_READWRITE);
 
+  /**
+   * KgxApplication:scrollback-lines:
+   *
+   * How many lines of scrollback #KgxTerminal should keep
+   *
+   * Bound to ‘scrollback-lines’ GSetting so changes persist
+   *
+   * Stability: Private
+   *
+   * Since: 0.4.0
+   */
+  pspecs[PROP_SCROLLBACK_LINES] =
+    g_param_spec_int64 ("scrollback-lines", "Scrollback Lines", "Size of the scrollback",
+                        G_MININT64, G_MAXINT64, 512,
+                        G_PARAM_READWRITE);
+
   g_object_class_install_properties (object_class, LAST_PROP, pspecs);
 }
 
+
 static void
 clear_watch (struct ProcessWatch *watch)
 {
   g_return_if_fail (watch != NULL);
 
   g_clear_pointer (&watch->process, kgx_process_unref);
-  g_clear_object (&watch->window);
+  g_clear_object (&watch->page);
 
   g_clear_pointer (&watch, g_free);
 }
 
+
 static void
 font_changed (GSettings      *settings,
               const char     *key,
@@ -499,8 +609,8 @@ font_changed (GSettings      *settings,
   g_object_notify_by_pspec (G_OBJECT (self), pspecs[PROP_FONT]);
 }
 
-static GOptionEntry entries[] =
-{
+
+static GOptionEntry entries[] = {
   {
     "version",
     0,
@@ -519,6 +629,15 @@ static GOptionEntry entries[] =
     NULL,
     NULL
   },
+  {
+    "tab",
+    0,
+    0,
+    G_OPTION_ARG_NONE,
+    NULL,
+    NULL,
+    NULL
+  },
   {
     "command",
     'e',
@@ -547,28 +666,136 @@ static GOptionEntry entries[] =
     N_("Wait until the child exits (TODO)"),
     NULL
   },
+  {
+    "title",
+    'T',
+    0,
+    G_OPTION_ARG_STRING,
+    NULL,
+    N_("Set the initial window title"),
+    NULL
+  },
+  {
+    "set-shell",
+    0,
+    0,
+    G_OPTION_ARG_STRING_ARRAY,
+    NULL,
+    N_("ADVANCED: Set the shell to launch"),
+    NULL
+  },
+  {
+    "set-scrollback",
+    0,
+    0,
+    G_OPTION_ARG_INT64,
+    NULL,
+    N_("ADVANCED: Set the scrollback length"),
+    NULL
+  },
   { NULL }
 };
 
+
+static void
+new_window_activated (GSimpleAction *action,
+                      GVariant      *parameter,
+                      gpointer       data)
+{
+  KgxApplication *self = KGX_APPLICATION (data);
+  guint32 timestamp = GDK_CURRENT_TIME;
+
+  kgx_application_add_terminal (self, NULL, timestamp, NULL, NULL, NULL);
+}
+
+
+static void
+new_tab_activated (GSimpleAction *action,
+                   GVariant      *parameter,
+                   gpointer       data)
+{
+  KgxApplication *self = KGX_APPLICATION (data);
+  guint32 timestamp = GDK_CURRENT_TIME;
+  g_autoptr (GFile) dir = NULL;
+  GtkWindow *window;
+
+  window = gtk_application_get_active_window (GTK_APPLICATION (self));
+  if (window) {
+    dir = kgx_window_get_working_dir (KGX_WINDOW (window));
+
+    kgx_application_add_terminal (self, KGX_WINDOW (window), timestamp, dir, NULL, NULL);
+  } else {
+    kgx_application_add_terminal (self, NULL, timestamp, NULL, NULL, NULL);
+  }
+}
+
+
 static void
 focus_activated (GSimpleAction *action,
                  GVariant      *parameter,
                  gpointer       data)
 {
   KgxApplication *self = KGX_APPLICATION (data);
-  GtkWindow *win;
+  GtkWidget *window;
+  KgxPages *pages;
+  KgxTab *page;
 
-  win = gtk_application_get_window_by_id (GTK_APPLICATION (self),
-                                          g_variant_get_uint32 (parameter));
+  page = kgx_application_lookup_page (self, g_variant_get_uint32 (parameter));
+  pages = kgx_tab_get_pages (page);
+  kgx_pages_focus_page (pages, page);
+  window = gtk_widget_get_toplevel (GTK_WIDGET (pages));
 
-  gtk_window_present_with_time (win, GDK_CURRENT_TIME);
+  gtk_window_present_with_time (GTK_WINDOW (window), GDK_CURRENT_TIME);
 }
 
-static GActionEntry app_entries[] =
+
+static void
+zoom_out_activated (GSimpleAction *action,
+                    GVariant      *parameter,
+                    gpointer       data)
 {
-  { "focus-window", focus_activated, "u", NULL, NULL },
+  KgxApplication *self = KGX_APPLICATION (data);
+
+  if (self->scale < 0.1) {
+    return;
+  }
+
+  kgx_application_set_scale (self, self->scale - 0.1);
+}
+
+
+static void
+zoom_normal_activated (GSimpleAction *action,
+                       GVariant      *parameter,
+                       gpointer       data)
+{
+  KgxApplication *self = KGX_APPLICATION (data);
+
+  kgx_application_set_scale (self, 1.0);
+}
+
+
+static void
+zoom_in_activated (GSimpleAction *action,
+                   GVariant      *parameter,
+                   gpointer       data)
+{
+  KgxApplication *self = KGX_APPLICATION (data);
+
+  kgx_application_set_scale (self, self->scale + 0.1);
+}
+
+
+static GActionEntry app_entries[] = {
+  { "new-window", new_window_activated, NULL, NULL, NULL },
+  { "new-tab", new_tab_activated, NULL, NULL, NULL },
+  { "focus-page", focus_activated, "u", NULL, NULL },
+  { "zoom-out", zoom_out_activated, NULL, NULL, NULL },
+  { "zoom-normal", zoom_normal_activated, NULL, NULL, NULL },
+  { "zoom-in", zoom_in_activated, NULL, NULL, NULL },
 };
 
+
 static void
 kgx_application_init (KgxApplication *self)
 {
@@ -594,37 +821,43 @@ kgx_application_init (KgxApplication *self)
                                     NULL,
                                     NULL,
                                     (GDestroyNotify) clear_watch);
+  self->pages = g_tree_new_full (kgx_pid_cmp,
+                                 NULL,
+                                 NULL,
+                                 (GDestroyNotify) g_object_unref);
 
   self->active = 0;
   self->timeout = 0;
 }
 
-#if HAS_GTOP
+
 /**
  * kgx_application_add_watch:
  * @self: the #KgxApplication
  * @pid: the shell process to watch
- * @window: the window the shell is running in
- * 
- * registers a new shell process with the pid watcher
+ * @page: the #KgxTab the shell is running in
+ *
+ * Registers a new shell process with the pid watcher
+ *
+ * Since: 0.3.0
  */
 void
 kgx_application_add_watch (KgxApplication *self,
                            GPid            pid,
-                           KgxWindow      *window)
+                           KgxTab        *page)
 {
   struct ProcessWatch *watch;
 
   g_return_if_fail (KGX_IS_APPLICATION (self));
-  g_return_if_fail (KGX_IS_WINDOW (window));
+  g_return_if_fail (KGX_IS_TAB (page));
 
   watch = g_new0 (struct ProcessWatch, 1);
   watch->process = kgx_process_new (pid);
-  watch->window = g_object_ref (window);
+  watch->page = g_object_ref (page);
 
   g_debug ("Started watching %i", pid);
 
-  g_return_if_fail (KGX_IS_WINDOW (watch->window));
+  g_return_if_fail (KGX_IS_TAB (watch->page));
 
   g_tree_insert (self->watching, GINT_TO_POINTER (pid), watch);
 }
@@ -633,7 +866,7 @@ kgx_application_add_watch (KgxApplication *self,
  * kgx_application_remove_watch:
  * @self: the #KgxApplication
  * @pid: the shell process to stop watch watching
- * 
+ *
  * unregisters the shell with #GPid pid
  */
 void
@@ -649,7 +882,7 @@ kgx_application_remove_watch (KgxApplication *self,
     g_warning ("Unknown process %i", pid);
   }
 }
-#endif
+
 
 /**
  * kgx_application_get_font:
@@ -673,6 +906,7 @@ kgx_application_get_font (KgxApplication *self)
   return pango_font_description_from_string (font);
 }
 
+
 /**
  * kgx_application_push_active:
  * @self: the #KgxApplication
@@ -695,6 +929,7 @@ kgx_application_push_active (KgxApplication *self)
   }
 }
 
+
 /**
  * kgx_application_pop_active:
  * @self: the #KgxApplication
@@ -715,4 +950,145 @@ kgx_application_pop_active (KgxApplication *self)
   } else {
     set_watcher (self, TRUE);
   }
-} 
\ No newline at end of file
+}
+
+
+/**
+ * kgx_application_add_page:
+ * @self: the instance to look for @id in
+ * @page: the page to add
+ *
+ * Register a new #KgxTab with @self
+ */
+void
+kgx_application_add_page (KgxApplication *self,
+                          KgxTab        *page)
+{
+  guint id = 0;
+
+  g_return_if_fail (KGX_IS_APPLICATION (self));
+  g_return_if_fail (KGX_IS_TAB (page));
+
+  id = kgx_tab_get_id (page);
+
+  g_tree_insert (self->pages, GINT_TO_POINTER (id), g_object_ref (page));
+}
+
+
+/**
+ * kgx_application_lookup_page:
+ * @self: the instance to look for @id in
+ * @id: the page id to look for
+ *
+ * Try and find a #KgxTab with @id in @self
+ *
+ * Returns: (transfer none) (nullable): the found #KgxTab or %NULL
+ */
+KgxTab *
+kgx_application_lookup_page (KgxApplication *self,
+                             guint           id)
+{
+  g_return_val_if_fail (KGX_IS_APPLICATION (self), NULL);
+
+  return g_tree_lookup (self->pages, GUINT_TO_POINTER (id));
+}
+
+
+static void
+started (GObject      *src,
+         GAsyncResult *res,
+         gpointer      app)
+{
+  g_autoptr (GError) error = NULL;
+  KgxTab *page = KGX_TAB (src);
+  GPid pid;
+
+  pid = kgx_tab_start_finish (page, res, &error);
+
+  if (error) {
+    g_warning ("Failed to start %s: %s",
+               G_OBJECT_TYPE_NAME (src),
+               error->message);
+
+    return;
+  }
+
+  kgx_application_add_watch (KGX_APPLICATION (app), pid, page);
+}
+
+
+KgxTab *
+kgx_application_add_terminal (KgxApplication *self,
+                              KgxWindow      *existing_window,
+                              guint32         timestamp,
+                              GFile          *working_directory,
+                              const char     *command,
+                              const char     *title)
+{
+  g_autofree char *user_shell = vte_get_user_shell ();
+  g_autofree char *directory = NULL;
+  g_autoptr (GError) error = NULL;
+  g_auto (GStrv) shell = NULL;
+  g_auto (GStrv) custom_shell = NULL;
+  GtkWindow *window;
+  GtkWidget *tab;
+  KgxPages *pages;
+
+  g_shell_parse_argv (command ? command : user_shell, NULL, &shell, &error);
+
+  if (error) {
+    g_warning ("Failed to parse “%s” as a command",
+               command ? command : user_shell);
+    shell = NULL;
+    g_clear_error (&error);
+  }
+
+  if (G_LIKELY (command == NULL)) {
+    custom_shell = g_settings_get_strv (self->settings, "shell");
+
+    if (g_strv_length (custom_shell) > 0) {
+      shell = g_steal_pointer (&custom_shell);
+    }
+  }
+
+  /* We should probably do something other than /bin/sh  */
+  if (shell == NULL) {
+    shell = g_new0 (char *, 2);
+    shell[0] = g_strdup ("/bin/sh");
+    shell[1] = NULL;
+    g_warning ("Defaulting to “%s”", shell[0]);
+  }
+
+  if (working_directory) {
+    directory = g_file_get_path (working_directory);
+  } else {
+    directory = g_strdup (g_get_home_dir ());
+  }
+
+  tab = g_object_new (KGX_TYPE_SIMPLE_TAB,
+                      "application", self,
+                      "visible", TRUE,
+                      "initial-work-dir", directory,
+                      "command", shell,
+                      "tab-title", title,
+                      "close-on-quit", command == NULL,
+                      NULL);
+  kgx_tab_start (KGX_TAB (tab), started, self);
+
+  if (existing_window) {
+    window = GTK_WINDOW (existing_window);
+  } else {
+    window = g_object_new (KGX_TYPE_WINDOW,
+                           "application", self,
+                           NULL);
+  }
+
+  pages = kgx_window_get_pages (KGX_WINDOW (window));
+
+  kgx_pages_add_page (pages, KGX_TAB (tab));
+  kgx_pages_focus_page (pages, KGX_TAB (tab));
+
+  gtk_window_present_with_time (window, timestamp);
+
+  return KGX_TAB (tab);
+}
